from . import functional as F
from .base import BaseTransform

try:
    from scipy.special import expit
except ModuleNotFoundError:
    # fake expit if scipy is not installed
    import numpy as np
    expit = lambda x: 1 / (1 + np.exp(-x))

__all__ = [
    "ToTensor",
    "ToInt",
    "ZeroMean",
    "CutWaveform",
    "SoftClip",
]

class ToTensor(BaseTransform):
    """Converts a numpy.ndarray C x S) to a torch.FloatTensor of shape (C x S).

    """

    def __call__(self, wav):
        return F._to_tensor(wav)


class ToInt(BaseTransform):
    """Convert a label to int based on the given lookup table.

    Args:
        lookup (dict): Lookup table to convert a label to int.

    """

    def __init__(self, lookup):
        if isinstance(lookup, dict):
            self.lookup = lookup
        else:
            raise ValueError("Lookup table needs to be a dictionary.")
        if any([not isinstance(val, int) for val in self.lookup.values()]):
            raise ValueError("Values of the lookup table need to be Int.")

    def __call__(self, label):
        return self.lookup[label]


class SoftClip(object):
    """Soft clip input to compress large amplitude signals.

    """

    def __init__(self, scale=1):
        super(SoftClip, self).__init__()
        self.scale = scale

    def __call__(self, wav):
        return expit(wav * self.scale)


class ZeroMean(BaseTransform):
    """Remove mean from each waveforms.

    """

    def __call__(self, wav):
        wav = wav.astype(float).T
        wav -= wav.mean(axis=0)
        return wav.T


class CutWaveform(BaseTransform):
    """Cut a portion of the input waveform.

    """

    def __init__(self, samplestart, sampleend):
        self.start = int(samplestart)
        self.end = int(sampleend)

    def __call__(self, wav):
        return wav[:, self.start:self.end]



import numpy as np

from .base import PathDataset

__all__ = [
    'FileDataset',
    'DatasetArray',
]


class FileDataset(PathDataset):
    """An abstract class representing a Dataset in a file.

    Args:
        path (object): File of the file.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (list): List of samples in the dataset.
        targets (list): List of targets in teh dataset.

    """

    def is_valid(self):
        return self.root.is_file()

    def handle_invalid(self):
        raise ValueError(f"{self.root} is not a file.")


class DatasetArray(FileDataset):
    """A generic data loader for ``.npy`` file where samples are arranged is
    the following way: ::

        array = [
            [sample0, target0],
            [sample1, target1],
            ...
        ]

    where both samples and targets can be arrays.

    Args:
        path (object): Path to the ``.npy`` file.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples: List of samples in the dataset.
        targets: List of targets in teh dataset.

    """

    def build_dataset(self):
        """Returns samples and targets.

        """
        data = np.load(self.root, allow_pickle=True)
        return data[:, 0], data[:, 1]



import numpy as np
import torch

def _is_numpy_waveform(wav):
    return isinstance(wav, np.ndarray) and (wav.ndim in {1, 2})

def _to_tensor(wav):
    """Convert a ``numpy.ndarray`` to tensor.
    See ``ToTensor`` for more details.
    Args:
        wav (numpy.ndarray): Waveform to be converted to tensor.
    Returns:
        Tensor: Converted array.
    """

    if not(_is_numpy_waveform(wav)):
        raise TypeError(f"wav should be ndarray. Got {type(wav)}")

    if wav.ndim == 1:
        wav = wav[None, :]

    return torch.from_numpy(wav).float()



# TO-DO: need to add model_zoo utility and pretrained models.
import torch.nn as nn

__all__ = [
    'Cpic',
    'cpic64',
]

model_urls = {
    'cpic64': 'https://lijunzhu.info'
}

class Cpic(nn.Module):

    def __init__(self):
        super(Cpic, self).__init__()
        # 2000 -> 1000
        self.layer1 = nn.Sequential(
            nn.Conv1d(3, 16, kernel_size=5, stride=1, padding=26, bias=False),
            nn.BatchNorm1d(16),
            nn.ReLU(),
            # nn.Sigmoid(),
            nn.MaxPool1d(2)
        )

        # 1000 -> 500
        self.layer2 = nn.Sequential(
            nn.Conv1d(16, 32, kernel_size=5, stride=1, padding=2, bias=False),
            nn.BatchNorm1d(32),
            nn.ReLU(),
            # nn.Sigmoid(),
            nn.MaxPool1d(2)
        )

        # 500 -> 250
        self.layer3 = nn.Sequential(
            nn.Conv1d(32, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 250 -> 125
        self.layer4 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 126 -> 63
        self.layer5 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 63 -> 31
        self.layer6 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 31 -> 15
        self.layer7 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 15 -> 7
        self.layer8 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 7 -> 3
        self.layer9 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 3 -> 1
        self.layer10 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # 3 -> 2
        self.layer11 = nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(2)
        )

        # # 3 -> 2
        # self.layer11 = nn.Sequential(
        #     nn.Conv1d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
        #     nn.BatchNorm1d(64),
        #     nn.ReLU(),
        #     nn.MaxPool1d(2)
        # )

        self.fc = nn.Linear(64 * 1, 3)

    def forward(self, x):
        # print(x.size())
        out = self.layer1(x)
        # print(out.size())
        out = self.layer2(out)
        # print(out.size())
        out = self.layer3(out)
        # print(out.size())
        out = self.layer4(out)
        # print(out.size())
        out = self.layer5(out)
        # print(out.size())
        out = self.layer6(out)
        # print(out.size())
        out = self.layer7(out)
        # print(out.size())
        out = self.layer8(out)
        # print(out.size())
        out = self.layer9(out)
        # print(out.size())
        out = self.layer10(out)
        # print(out.size())
        out = self.layer11(out)
        # print(out.size())
        out = out.view(out.size(0), -1)
        # print(out.size())
        out = self.fc(out)

        return out

def cpic64(pretrained=False, **kwargs):
    r"""CPIC model architecture from the
    `"Deep learning for ..." <https://arxiv.org/abs/1901.06396>`_ paper.

    Args:
        pretrained (bool): If True, returns a model pre-trained on Wenchuan)
    """
    model = Cpic(**kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['cpic64']))
    return model



from .base import BaseDataset
from .base import is_dataset
from .base import PathDataset
from .dirs import DatasetArrayFolder
from .dirs import DatasetFolder
from .dirs import DirDataset
from .files import DatasetArray
from .files import FileDataset
from .packaged_datasets import Mariana
from .packaged_datasets import PackagedDataset
from .packaged_datasets import Wenchuan
from .sac import MarianaFromSource
from .utils import *

__all__ = (
    'BaseDataset',
    'PathDataset',
    'FileDataset',
    'DirDataset',
    'DatasetArray',
    'DatasetFolder',
    'DatasetArrayFolder',
    'PackagedDataset',
    'Wenchuan',
    'Mariana',
    'MarianaFromSource',
)



__all__ = [
    "MovingAverageMeter",
    "ExponentialMovingAverageMeter",
    "CumulativeMovingAverageMeter",
]

# -----------------------------------------------------------------------------
#
#   Moving average meters
#
# -----------------------------------------------------------------------------

class MovingAverageMeter(object):

    def __init__(self):
        self.val = None
        self.avg = None
        self.count = 0

    def reset(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError


class ExponentialMovingAverageMeter(MovingAverageMeter):
    """Computes and stores the exponential moving average and current value."""

    def __init__(self, alpha=0.5):
        super(ExponentialMovingAverageMeter, self).__init__()
        self.alpha = alpha
        self.reset()

    def reset(self):
        self.val = None
        self.avg = None
        self.count = 0

    def update(self, val, n=1):
        self.count += n
        self.val = val
        for i in range(n):
            if self.avg:
                self.avg = self.alpha * self.val + (1 - self.alpha) * self.avg
            else:
                self.avg = self.val

    def set_alpha(self, alpha):
        self.alpha = alpha


class CumulativeMovingAverageMeter(MovingAverageMeter):
    """Computes and stores the cumulative moving average and current value"""

    def __init__(self):
        super(CumulativeMovingAverageMeter, self).__init__()
        self.reset()

    def reset(self):
        self.val = None
        self.avg = None
        self.count = 0

    def update(self, val, n=1):
        if self.avg:
            previous_sum = self.avg * self.count
            self.avg = (previous_sum + n * val) / (self.count + n)
        else:
            self.avg = val

        self.val = val
        self.count += n



from pathlib import Path

from torch.utils import data

from .utils import create_npy

try:
    from tqdm import tqdm
except ModuleNotFoundError:
    # fake tqdm if it's not installed
    def tqdm(iterator):
        for i, item in enumerate(iterator):
            if i % 1000 == 0:
                print(f"{i} / {len(iterator)}")
            yield item

def is_dataset(obj):
    r"""Verfy if a object is ``dataset-like`` defined in
    :class:`torch.utils.data.Dataset`.

    Args:
        obj: Object to be determined.

    Returns:
        bool: True for ``dataset-like`` object, false otherwise.

    """
    return hasattr(obj, '__getitem__') and hasattr(obj, '__len__')


class BaseDataset(data.Dataset):
    r"""An abstract class representing a Dataset.

    All other datasets should subclass it. All subclasses should override
    ``build_dataset`` which construct the dataset-like object from root.

    Note:
        A dataset-like object has both ``__len__`` and ``__getitem__``
        impolemented.

    Args:
        root (object): Source of the dataset.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (dataset-like object): Dataset-like object for samples.
        targets (dataset-like object): Dataset-like object for targets.


    """

    _repr_indent = 4

    def __init__(self, root=None, sample_transform=None, target_transform=None):
        self.root = root

        if self.is_valid():
            self.samples, self.targets = self.build_dataset()
            # verify self.samples and self.targets are dataset-like object
            if not is_dataset(self.samples):
                raise ValueError("self.samples is not a dataset-like object")
            if not is_dataset(self.targets):
                raise ValueError("self.targets is not a dataset-like object")

            if len(self.samples) == len(self.targets):
                self.size = len(self.targets)
            else:
                raise ValueError("Samples and targets have different lengths.")
        else:
            self.handle_invalid()

        self.sample_transform = sample_transform
        self.target_transform = target_transform

    def is_valid(self):
        return self.root is not None

    def handle_invalid(self):
        self.size = 0

    def build_dataset(self):
        """Method to construct ``samples`` and ``targets`` from ``self.root``.

        Returns:
            Constructed dataset-like objects of samples and targets. They will
            be stored in ``self.samples`` and ``self.targets`` by ``__init__``.

        """
        raise NotImplementedError

    def export_dataset(self, path):
        path = Path(path)
        # get array shape
        samples_shape = (self.__len__(), ) + self.__getitem__(0)[0].shape
        targets_shape = (self.__len__(), ) + self.__getitem__(0)[1].shape
        # get array dtype
        samples_dtype = self.__getitem__(0)[0].dtype
        targets_dtype = self.__getitem__(0)[1].dtype
        # rseerve disk space
        fs = create_npy(path / 'samples.npy', samples_shape, samples_dtype)
        ft = create_npy(path / 'targets.npy', targets_shape, targets_dtype)

        # populate memmap numpy array
        for i in tqdm(range(self.__len__())):
            # add one item in the dataset
            fs[i] = self[i][0]
            ft[i] = self[i][1]

            # update disk files
            fs.flush()
            ft.flush()

        del fs
        del ft

    def __getitem__(self, index):
        """Indexing the dataset.

        Args:
            index (int): Index.

        Returns:
            Tuple of (samples, targets) combination.

        """
        sample = self.samples[index]
        target = self.targets[index]

        if self.sample_transform is not None:
            sample = self.sample_transform(sample)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target

    def __len__(self):
        return self.size

    def __repr__(self):
        head = "Dataset " + self.__class__.__name__
        body = ["Number of datapoints: {}".format(self.__len__())]
        if self.root is not None:
            body.append("Root location: {}".format(self.root))
        body += self.extra_repr().splitlines()
        if self.sample_transform is not None:
            body += self._format_transform_repr(self.sample_transform,
                                                "Sample transforms: ")
        if self.target_transform is not None:
            body += self._format_transform_repr(self.target_transform,
                                                "Target transforms: ")
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)

    @staticmethod
    def _format_transform_repr(transform, head):
        """Format transform representation for ``BaseDataset``.

        Args:
            transform (transform-like): Transform to be formated
            head (str): Formating string.

        Returns:
            str: Formated string.

        """
        lines = transform.__repr__().splitlines()
        return (["{}{}".format(head, lines[0])] +
                ["{}{}".format(" " * len(head), line) for line in lines[1:]])

    def extra_repr(self):
        return ""


class PathDataset(BaseDataset):
    """An abstract class representing a Dataset defined by a Path.

    Args:
        path (object): Path to the dataset.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (list): List of samples in the dataset.
        targets (list): List of targets in teh dataset.

    """

    def __init__(self, path, **kwargs):
        path = Path(path).resolve()    # make a Path object regardless of existence
        super().__init__(root=path, **kwargs)

    def is_valid(self):
        """Determine if the root path is valid.

        Other subclasses should overload this method if valid paths are defined
        differently.

        """

        return self.root.exists()

    def handle_invalid(self):
        raise ValueError(f"{self.root} is not a valid path.")



from .dirs import DatasetArrayFolder
from .utils import extract_bz2
from .utils import URL

class PackagedDataset(DatasetArrayFolder):
    """Packaged Dataset in .tar.bz2 foramt.

    Args:
        path (str): Root directory where ``dataset_name/samples.npy`
            and ``dataset_name/targets.npy`` exists.
        download (bool, optional): If True, downloads the dataset from internet
            and puts it in root directory. If dataset is already downloaded, it
            will not be downloaded again.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    """

    url = ''

    def __init__(self, download=False, **kwargs):
        # verify download flag
        if not isinstance(download, bool):
            raise ValueError("`download` needs to be True or False.")

        # verify if dataset is ready
        try:
            super().__init__(**kwargs)
        except ValueError:
            if download:
                url = URL(self.url)
                # download compressed file from source if not exists
                fpath = self.root / url.url_filename
                if not fpath.is_file():
                    url.download(self.root)
                # extract file under root directory
                print("Extracting dataset ...")
                extract_bz2(fpath, self.root)
                # try initiate DatasetArrayFolder again
                super().__init__(**kwargs)
            else:
                raise ValueError(f"{self.root} contains no valid dataset. "
                                 f"Consider set `download=True` and remove broken bz2 file.")


class Wenchuan(PackagedDataset):
    """`Wenchuan <https://arxiv.org/abs/1901.06396>`_ Dataset.

    Args:
        path (str): Root directory where ``wenchuan/samples.npy`
            and ``wenchuan/targets.npy`` exists.
        download (bool, optional): If True, downloads the dataset from internet
            and puts it in root directory. If dataset is already downloaded, it
            will not be downloaded again.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    """

    url = 'https://www.dropbox.com/s/enr75zt2ukx118r/wenchuan.tar.bz2?dl=1'


class Mariana(PackagedDataset):
    """`Mariana <https://arxiv.org/abs/1901.06396>`_ Dataset.

    Args:
        path (str): Root directory where ``mariana/samples.npy`
            and ``mariana/targets.npy`` exists.
        download (bool, optional): If True, downloads the dataset from internet
            and puts it in root directory. If dataset is already downloaded, it
            will not be downloaded again.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    """

    url = 'https://www.dropbox.com/s/6wqjhvk6hkxnuis/mariana.tar.bz2?dl=1'



import numpy as np

from . import utils
from .dirs import DatasetFolder
from .dirs import get_files_under_dir

def load_three_component_sac(sample):
    """Read a seismic frame given a sample from datasets.

    """
    path, starttime, endtime = sample
    return utils.read_frame_obspy(path, starttime=starttime, endtime=endtime)


class MarianaFromSource(DatasetFolder):
    """`Mariana <https://arxiv.org/abs/1901.06396>`_ Dataset stored in folder.

        root/.../event_p/sta-year_julday_hr_min_sec_msec-mag.BHX
        root/.../event_p/sta-year_julday_hr_min_sec_msec-mag.BHY
        root/.../event_p/sta-year_julday_hr_min_sec_msec-mag.BHZ

    Args:
        path (str): Root directory where ``wenchuan/processed/data.npy`
            and ``wenchuan/processed/labels.npy`` exists.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    """

    def __init__(self, **kwargs):
        self.raw_samples = None
        super().__init__(loader=load_three_component_sac, **kwargs)

    def build_dataset(self):
        """Construct samples list and targets list.

        """

        # get file patterns and arrival times
        files = get_files_under_dir(self.root, '**/*.BH*')
        patterns = [p.parent / (p.name.split('.BH')[0] + '.BH*') for p in files]
        arrivals = [p.name.split('-')[2].split('_') for p in patterns]
        arrivals = [f"{c[0]}-{c[1]}T{c[2]}:{c[3]}:{c[4]}" for c in arrivals]

        # construct samples and labels
        samples = []
        targets = []
        for i in range(len(patterns)):
            starttime = utils.UTCDateTime(arrivals[i]) - 5
            endtime = utils.UTCDateTime(arrivals[i]) + 15
            filename = str(patterns[i])

            # phase
            target = {
                'event_p': 1,
                'event_s': 2,
            }.get(patterns[i].parent.name)

            samples.append((filename, starttime, endtime))
            targets.append(target)

            # noise
            offset = -75 if target==1 else 65
            starttime += offset
            endtime += offset
            target = 0

            samples.append((filename, starttime, endtime))
            targets.append(target)

        self.raw_samples = samples
        samples = self.FilesLoader(samples, self.loader)
        targets = np.array(targets)
        return samples, targets



import math
import os
import tarfile
from pathlib import Path
from urllib import request

import numpy as np
from torch.utils.model_zoo import tqdm

try:
    from obspy import read
    from obspy import UTCDateTime
    has_obspy = True
except ModuleNotFoundError:
    has_obspy = False

################################################################################
#
#                       ObsPy related reading and loading
#
################################################################################
def stream2array(st):
    """Convert seismic frame from obspy.Stream to numpy.ndarray.

    """
    if has_obspy:
        return np.stack([tr.data for tr in st])
    else:
        raise ModuleNotFoundError("Consider installing ObsPy for seismic I/O.")


def read_frame_obspy(path, **kwargs):
    """Read a seismic frame using ObsPy read.

    Args:
        path (path): Path to seismic files (SAC, mseed, etc.).
        starttime (UTCDateTime, optional): Frame starting time.
        endtime (UTCDateTime, optional): Frame ending time.

    Returns:
        Cutted frame numpy array of single or multiple component seismogram.

    """
    if has_obspy:
        return stream2array(read(path, **kwargs))
    else:
        raise ModuleNotFoundError("Consider installing ObsPy for seismic I/O.")


################################################################################
#
#                   Read and Write with Memory Constraints
#
################################################################################

# dynamically change this limit by Python code.
MEMORY_LIMIT =  2 * 1024 ** 3        # 2 GB limit

def over_memory_limit(path):
    return os.stat(path).st_size > MEMORY_LIMIT

def set_memory_limit(limit):
    global MEMORY_LIMIT
    MEMORY_LIMIT = limit

def get_memory_limit():
    return MEMORY_LIMIT

def load_npy(path, memory_limit=None):
    default_memory_limit = get_memory_limit()

    if memory_limit:
        set_memory_limit(memory_limit)
    print(f"Current memory limit is {sizeof_fmt(get_memory_limit())}")

    if over_memory_limit(path):
        print(f"Loading memory map of {path} into memory")
        data = np.load(path, mmap_mode='r')
    else:
        print(f"Loading {path} directly into memory")
        data = np.load(path)
    set_memory_limit(default_memory_limit)

    return data

def create_npy(path, shape, dtype=np.float32):
    return np.lib.format.open_memmap(path, mode='w+', dtype=dtype, shape=shape)

################################################################################
#
#                           Utilities for URLs
#
################################################################################

def test_url(url):
    try:
        with request.urlopen(url) as req:
            return req
    except:
        return None

def gen_bar_update():
    pbar = tqdm(total=None)

    def bar_update(count, block_size, total_size):
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)

    return bar_update

def sizeof_fmt(num, suffix='B'):
    """Get human-readable file size

    Args:
        num (int): Number of unit size.
        suffix (str): Unit of size (default: B).

    """
    mag = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, mag)
    if mag > 7:
        return f"{val:.1f}Y{suffix}"
    else:
        return f"{val:3.1f}{['','K','M','G','T','P','E','Z'][mag]}{suffix}"


class URL(object):

    def __init__(self, url):
        self.url = url
        req = test_url(url)
        if req:
            self.req = req
            self.size = req.length
            self.url_filename = self.get_filename()
        else:
            raise ValueError(f"{url} is not a valid URL or unreachable.")

    def get_filename(self):
        return [a.split('=')[1] for a in self.req.info()['content-disposition'].split('; ')
                    if a.startswith('filename=')][0].strip('"')

    def __repr__(self):
        return f"{self.url_filename}, {sizeof_fmt(self.size)}, from <{self.url}>"

    def download(self, root, filename=None):
        root = Path(root)
        root.mkdir(parents=True, exist_ok=True)    # mkdir if not exists)
        if not filename:
            filename = self.url_filename
        fpath = root / filename

        print(f"Downloading {self.url} to {fpath}")
        request.urlretrieve(self.url, fpath, reporthook=gen_bar_update())


################################################################################
#
#                           Utilities for Bz2
#
################################################################################

def extract_bz2(infile, outdir='.'):
    with tarfile.open(infile, 'r:bz2') as tar:
        tar.extractall(outdir)



from pathlib import Path
import shutil
import time
import numpy as np
import torch

from .utils import CumulativeMovingAverageMeter

def get_torch_device():
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def model_on_device(model, device):
    return torch.nn.DataParallel(model.to(device))

def model_off_device(model):
    return model.module.to(torch.device('cpu')).state_dict()

def generate_tmp_name(prefix):
    return f"{prefix}_{np.floor(time.time()):.0f}"


def rm_dir_content(path):
    path = Path(path)
    shutil.rmtree(path)

def rm_content(path):
    for content in Path(path).glob('*'):
        shutil.rmtree(content)

def save_checkpoint(state, is_best=False, filename='checkpoint.pth.tar'):
    """Save checkpoint to pth.tar file"""

    filename = Path(filename)
    print(f"=> saving checkpoint")
    torch.save(state, filename)
    print(f"=> checkpoint saved to '{filename}'")
    if is_best:
        shutil.copyfile(filename, filename.parent / 'model_best.pth.tar')
        print("=> best model updated.")

def load_checkpoint(filename):
    """Load checkpoint from pth.tar file"""

    if Path(filename).is_file():
        print(f"=> loading checkpoint '{filename}'")

        checkpoint = torch.load(filename)
        print(f"=> loaded checkpoint '{filename}' "
              f"(epoch {checkpoint['epoch']})")

        return checkpoint
    else:
        print(f"=> no checkpoint found at '{filename}'")

        return None

def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(float(correct_k.mul_(100.0 / batch_size)))
    return res

def train(model, train_loader, criterion, optimizer, epoch, print_freq):
    # performance metrics
    batch_time = CumulativeMovingAverageMeter()
    data_time = CumulativeMovingAverageMeter()
    losses = CumulativeMovingAverageMeter()
    top1 = CumulativeMovingAverageMeter()

    # swtich to train mode
    model.train()

    end = time.time()
    optimizer.zero_grad()
    for i, (samples, target) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        # model target to cuda
        if next(model.parameters()).is_cuda:
            target = target.cuda(non_blocking=True)

        output = model(samples)
        loss = criterion(output, target)

        # measure accuracy and record loss
        prec1 = accuracy(output, target)
        losses.update(loss.item(), samples.size(0))
        top1.update(prec1[0], samples.size(0))

        # compute gradient and do SGD step
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if print_freq:
            if (i % print_freq == 0):
                print(f'Epoch: [{epoch}][{i} / {len(train_loader)}]\t'
                      f'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                      f'Data {data_time.val:.3f} ({data_time.avg:.3f})\t'
                      f'Loss {losses.val:.4f} ({losses.avg:.4f})\t'
                      f'Prec@1 {top1.val:.3f} ({top1.avg:.3f})')

    return top1.avg, losses.avg


def validate(model, val_loader, criterion, print_freq=None):
    batch_time = CumulativeMovingAverageMeter()
    losses = CumulativeMovingAverageMeter()
    top1 = CumulativeMovingAverageMeter()

    # switch to evaluate model
    model.eval()

    with torch.no_grad():
        end = time.time()
        for i, (samples, target) in enumerate(val_loader):
            if next(model.parameters()).is_cuda:
                target = target.cuda(non_blocking=True)

            output = model(samples)
            loss = criterion(output, target)

            # measure accuracy and record loss
            prec1 = accuracy(output, target)
            losses.update(loss.item(), samples.size(0))
            top1.update(prec1[0], samples.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if print_freq:
                if i % print_freq == 0:
                    print(f"Test [{i}/{len(val_loader)}]\t"
                          f"Loss {losses.val:.4f} ({losses.avg:.4f})\t"
                          f"Prec@1 {top1.val:.3f} ({top1.avg:.3f})")

    print(f" * Prec@1 {top1.avg:.3f}")

    return top1.avg, losses.avg



from torch import optim

from . import functional as F

class Trainer(object):
    """Mega class for training process.

    """
    def __init__(self, model_gen, criterion, lr=0.1):
        # default device
        self.device = F.get_torch_device()

        # model
        self.model_gen = model_gen
        self.model = None

        # optimizer
        self.criterion = criterion
        self.lr = lr
        self.optimizer = None
        self.scheduler = None

        # training process
        self.start_epoch = 0
        self.end_epoch = None

        # results
        self.best_acc = None
        self.best_model = None
        self.train_loss = None
        self.train_acc = None
        self.val_loss = None
        self.val_acc = None

        # reset
        self.reset()


    def reset(self):
        self.model = self.model_gen()
        self.model = F.model_on_device(self.model, self.device)

        self.reset_optimizer()
        self.reset_scheduler()

        self._reset_results()

    def reset_optimizer(self):
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.lr)

    def reset_scheduler(self):
        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optimizer,
                                                              verbose=True)

    def _update_scheduler(self):
        self.scheduler.step(self.val_loss[-1])


    def _reset_results(self):
        self.best_acc = 0.
        self.best_model = None
        self.train_loss = []
        self.train_acc = []
        self.val_loss = []
        self.val_acc = []

    def update_train_results(self, acc, loss):
        self.train_acc.append(acc)
        self.train_loss.append(loss)

    def update_val_results(self, acc, loss):
        self.val_acc.append(acc)
        self.val_loss.append(loss)

    def validate(self, loader, print_freq=None):
        """Validate the model on a given datset using the dataloader provided.

        """

        return F.validate(self.model, loader, self.criterion, print_freq=print_freq)

    def train_one_epoch(self, loader, epoch, print_freq=None):
        return F.train(self.model, loader, self.criterion, self.optimizer,
                       epoch, print_freq=print_freq)

    def train(self, train_loader, val_loader, epochs=100, print_freq=None):
        """Train the model on a given datset using the dataloader provided.

        """
        start_epoch = 0
        end_epoch = epochs

        # record results for initial model
        print("Validation on training set.")
        acc, loss = self.validate(train_loader)
        self.update_train_results(acc, loss)
        print("Validation on valiation set.")
        acc, loss = self.validate(val_loader)
        self.update_val_results(acc, loss)

        # training loop
        print("Start training ...")
        for epoch in range(start_epoch, end_epoch):
            # update learning rate
            self._update_scheduler()

            # train model for one epoch
            acc, loss = self.train_one_epoch(train_loader, epoch, print_freq=print_freq)
            self.update_train_results(acc, loss)

            # validate model for current epoch
            acc, loss = self.validate(val_loader)
            self.update_val_results(acc, loss)

            # preserve best model and accuracy
            is_best = self.val_acc[-1] > self.best_acc
            self.best_acc = max(self.val_acc[-1], self.best_acc)
        print(f"Training fisihed. Best accuracy is {self.best_acc}")



def is_transform(obj):
    """Verfy if a object is a ``transform-like`` object.

    Args:
        obj: Object to be determined.

    Returns:
        bool: True for ``transform-like`` object, false otherwise.

    """
    return hasattr(obj, '__call__')

class BaseTransform(object):
    """An abstract class representing a Transform.

    All other transform should subclass it. All subclasses should override
    ``__call__`` which performs the transform.

    Note:
        A transform-like object has ``__call__`` implmented. Typical
        transform-like objects include python functions and methods.

    """

    def __call__(self, data):
        raise NotImplementedError

    def __repr__(self):
        head = self.__class__.__name__
        content = [f"{key} = {val}" for key, val in self.__dict__.items()]
        body = ", ".join(content)
        return f"{head}({body})"


class Compose(BaseTransform):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.ZeroMean(),
        >>>     transforms.ToTensor(),
        >>> ])

    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, wav):
        for t in self.transforms:
            wav = t(wav)
        return wav

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string



import numpy as np

from . import utils
from .base import PathDataset

__all__ = [
    'DirDataset',
    'DatasetArrayFolder',
    'DatasetFolder',
]

def get_files_under_dir(path, pattern):
    return [p for p in path.glob(pattern) if p.is_file()]


class DirDataset(PathDataset):
    """An abstract class representing a Dataset in a directory.

    Args:
        path (object): Path to the directory.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (list): List of samples in the dataset.
        targets (list): List of targets in teh dataset.

    """

    def is_valid(self):
        """Determine if root is a valid directory.

        """
        path_exists = super().is_valid()
        valid_dir = self.root.is_dir()
        return path_exists and valid_dir

    def handle_invalid(self):
        raise ValueError(f"{self.root} is not a valid directory.")


class DatasetArrayFolder(DirDataset):
    """A generic data loader for a folder of ``.npy`` files where samples are
    arranged in the following way: ::

        root/samples.npy: each row is a sample
        root/targets.npy: each row is a target

    where both samples and targets can be arrays.

    Args:
        path (object): Path to the dataset folder.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (list): List of samples in the dataset.
        targets (list): List of targets in the dataset.

    """

    def is_valid(self):
        """Determine if root is a valid array folder.

        """
        valid_dir = super().is_valid()
        has_samples = (self.root / 'samples.npy').exists()
        has_targets = (self.root / 'targets.npy').exists()
        return  valid_dir and has_samples and has_targets

    def handle_invalid(self):
        raise ValueError(f"{self.root} is not a valid array folder. Requires samples.npy and targets.npy.")

    def build_dataset(self):
        """Returns samples and targets.

        """
        samples_path = self.root / 'samples.npy'
        targets_path = self.root / 'targets.npy'

        samples = utils.load_npy(samples_path)
        targets = utils.load_npy(targets_path)

        return samples, targets


class DatasetFolder(DirDataset):
    """A generic data loader for a folder where samples are arranged in the
    following way: ::

        root/.../class_x.xxx
        root/.../class_x.sdf3
        root/.../class_x.asd932

        root/.../class_y.yyy
        root/.../class_y.as4h
        root/.../blass_y.jlk2

    Args:
        path (path): Path to the dataset.
        loader (callable): Function that load one sample from a file.
        sample_transform (callable, optional): A function/transform that takes
            a sample and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            a target and transform it.

    Attributes:
        samples (list): List of samples in the dataset.
        targets (list): List of targets in teh dataset.


    """

    class FilesLoader(object):
        """A dataset-like class for loading a list of files given a loader.

        Args:
            files (list): List of file paths
            loader (callable): Function that load one file.

        """

        def __init__(self, files, loader):
            self.files = files
            self.loader = loader

        def __getitem__(self, index):
            return self.loader(self.files[index])

        def __len__(self):
            return len(self.files)

    def __init__(self, loader, **kwargs):
        # TO-DO: check if root directory is empty
        self.loader = loader
        super().__init__(**kwargs)

    def build_dataset(self):
        """Return samples and targets.

        """
        files = get_files_under_dir(self.root, '**/*')
        labels = [p.name.split('.')[0] for p in files]
        samples = self.FilesLoader(files, self.loader)

        return samples, labels
